@@ -5,67 +5,61 @@ from efficientnet_pytorch import EfficientNet
 
 
 # Utility Functions for the model
-
-# Hook functions to get values of intermediate layers for cross connection
-encoder_out = []
-def hook(_, input, output):
-	global encoder_out
-	encoder_out.append(output) # stores values of each layers in encoder_out
-def inhook(_,inp,output):
-	global encoder_out
-	encoder_out.append(inp[0])
-
-# Initialize encoder to get shapes of selected layers
-shapes = []
-def init_hook(model,device,concat_input):
-	global shapes, encoder_out
-	shapes = []
-	encoder_out = []
-
-	if concat_input:
-		model._conv_stem.register_forward_hook(inhook)
-
-	hooks = []
-	for i in range(len(model._blocks)):
-		hooks.append(model._blocks[i].register_forward_hook(hook)) #register hooks
-	
-	image = torch.rand([1,3,576,576]) # doesn't matter if input size is different from this during training.
-	image = image.to(device)
-	out = model(image) # generate hook values to get shapes
-
-	shapes = [encoder_out[i].shape for i in range(len(encoder_out)-1) if encoder_out[i].shape[2] != encoder_out[i+1].shape[2]]
-
-	if concat_input:
-		indices = [i-1 for i in range(1,len(encoder_out)-1) if encoder_out[i].shape[2] != encoder_out[i+1].shape[2]]
-		shapes.append(encoder_out[-1].shape)
-		indices.append(len(encoder_out[1:])-1)
-	else:
-		indices = [i for i in range(0,len(encoder_out)-1) if encoder_out[i].shape[2] != encoder_out[i+1].shape[2]]
-		shapes.append(encoder_out[-1].shape)
-		indices.append(len(encoder_out)-1)
-
-	for i,_hook in enumerate(hooks):
-		if i not in indices:
-			_hook.remove()
-
-	shapes = shapes[::-1] 
-	encoder_out=[]
-
 def double_conv(in_,out_,drop): # Double convolution layer for decoder 
 	conv = nn.Sequential(
 		nn.Conv2d(in_,out_,kernel_size=3,padding=(1,1)),
-		nn.BatchNorm2d(out_),
 		nn.ReLU(inplace=True),
 		nn.Conv2d(out_,out_,kernel_size=3,padding=(1,1)),
-		nn.BatchNorm2d(out_),
 		nn.ReLU(inplace=True),
-		nn.Dropout(drop)
+        nn.Dropout(drop)
 		)
 	return conv
 
+def crop(tensor,target_tensor): # Crop tensor to target tensor size
+	target_shape = target_tensor.shape[2]
+	return T.CenterCrop(target_shape)(tensor)
+
+
+# Hook functions to get values of intermediate layers for cross connection
+hook_values = []
+def hook(_, input, output):
+	global hook_values
+	hook_values.append(output) # stores values of each layers in hook_values
+
+indices = []
+shapes = []
+def init_hook(model,device):
+	global shapes, indices, hook_values
+
+	for i in range(len(model._blocks)):
+		model._blocks[i].register_forward_hook(hook) #register hooks
+	
+	image = torch.rand([1,3,576,576])
+	image = image.to(device)
+	out = model(image) # generate hook values to get shapes
+	
+	shape = [i.shape for i in hook_values] # get shape of all layers
+	
+	for i in range(len(shape)-1):
+		if shape[i][2]!=shape[i+1][2]: # get indices of layers only where output dimension change
+			indices.append(i)
+	indices.append(len(shape)-1) # get last layer index
+	
+	shapes = [shape[i] for i in indices] # get shapes of required layers
+	shapes = shapes[::-1]  
+
+encoder_out = []
+def epoch_hook(model, image):
+	global encoder_out, indices, hook_values
+	hook_values = []
+
+	out = model(image) # generate layer outputs with current image
+	encoder_out = [hook_values[i] for i in indices] # get layer outputs for selected indices
+
+
 class EffUNet(nn.Module):
 
-	def __init__(self,model='b0',out_channels=1,dropout=0.1,freeze_backbone=True,pretrained=True,concat_input=True,device='cuda'):
+	def __init__(self,model='b0',out_channels=2,dropout=0.1,freeze_backbone=True,pretrained=True,device='cuda',num_gpu=1):
 		super(EffUNet,self).__init__()
 		global layers, shapes
 
@@ -76,7 +70,7 @@ class EffUNet(nn.Module):
 		else:
 			self.encoder = EfficientNet.from_name(f'efficientnet-{model}')
 
-		# Disable non required layers by replacing them with identity to save time and memory
+		# Disable non required layers by replacing them with identity
 		self.encoder._conv_head=torch.nn.Identity()
 		self.encoder._bn1=torch.nn.Identity()
 		self.encoder._avg_pooling=torch.nn.Identity()
@@ -84,14 +78,13 @@ class EffUNet(nn.Module):
 		self.encoder._fc=torch.nn.Identity()
 		self.encoder._swish=torch.nn.Identity()
 
-		if not concat_input:
-			self.encoder._conv_stem.stride=(1,1)
-
 		if isinstance(device, str):
 			self.device = torch.device(device)
 		else:
 			self.device = device
 		self.encoder.to(self.device)
+		self.encoder._conv_stem.stride=1 # change stride of first layer from 2 to 1 to increase o/p size
+		self.encoder._conv_stem.kernel_size=(1,1) # 
 
 		# freeze encoder
 		if freeze_backbone:
@@ -99,7 +92,7 @@ class EffUNet(nn.Module):
 				param.requires_grad = False
 
 		# register hooks & get shapes
-		init_hook(self.encoder,self.device,concat_input)
+		init_hook(self.encoder,self.device)
 
 		# Building decoder
 		self.decoder = torch.nn.modules.container.ModuleList()
@@ -111,17 +104,29 @@ class EffUNet(nn.Module):
 		#output layer
 		self.out = nn.Conv2d(shapes[-1][1],out_channels,kernel_size=1).to(self.device)
 
+		# Handling multiple GPUs
+		if num_gpu>1 and device=='cuda':
+			self.encoder = nn.DataParallel(self.encoder)
+
 	def forward(self, image):
-		global layers,encoder_out
+		global layers
+		
+		h=image.shape[2]
+		w=image.shape[3]
+		if h%8!=0 or w%8!=0:
+			new_h = round(h/8)*8
+			new_w = round(w/8)*8
+			image =  T.Resize((new_h,new_w))(image)
 
 		# Encoder
-		encoder_out=[]
-		self.encoder(image) # required outputs accumulate in "encoder_out"
+		epoch_hook(self.encoder, image) # required outputs accumulate in "encoder_out"
+
 		#Decoder
 		x = encoder_out.pop()
 		for i in range(len(self.decoder)):
 			x = self.decoder[i][0](x) # conv transpose
 			prev = encoder_out.pop()
+			prev = crop(prev,x) # croping for cross connection
 			prev = torch.cat([x,prev],axis=1) # concatenating 
 			x = self.decoder[i][1](prev) # double conv
 		
@@ -129,11 +134,8 @@ class EffUNet(nn.Module):
 		x = self.out(x)
 		return x
 
-# img = torch.rand([1,3,576,576]).cuda()
-# from time import time
-# start=time()
+# img = torch.rand([1,3,512,512]).cuda()
 # model = EffUNet()
-# print(time()-start)
 # # print(model)
 # out = model(img)
 # print(out.shape)
\ No newline at end of file
