@@ -12,11 +12,9 @@ def double_conv(in_,out_,drop):
 		)
 	return conv
 
-def crop(tensor,target_tensor): 
+def crop(tensor,target_tensor):
 	target_shape = target_tensor.shape[2]
-	tensor_size = tensor.size()[2]
-	delta = (target_shape-tensor_size)//2
-	return tensor[:,:,delta:tensor_size-delta,delta:tensor_size-delta]
+	return T.CenterCrop(target_shape)(tensor)
 
 class UNet(nn.Module):
 	def __init__(self,dropout=0.1):
